#include "ShareFileTransfer.h"
#include "ShareWindow.h"
#include <kernel/fs_attr.h>
#include <interface/ListView.h>
#include <interface/Region.h>
#include <storage/Mime.h>
#include <storage/NodeInfo.h>

#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"

#include "Colors.h"
#include "ShareUtils.h"

const rgb_color PalePink = {255, 233, 233};
const rgb_color PaleBlue = {233, 233, 255};

const rgb_color BrightPink = {255, 200, 200};
const rgb_color BrightBlue = {200, 200, 255};

ShareFileTransfer ::
ShareFileTransfer(const BDirectory & fileDir, const char * localSessionID)
   : _mtt(NULL), _dir(fileDir), _currentFileBytesDone(-1), _currentFileSize(-1),
     _onConnect(TRANSFER_COMMAND_CONNECTED_TO_PEER), _onDisconnect(TRANSFER_COMMAND_DISCONNECTED_FROM_PEER),
     _onEmpty(TRANSFER_OUTPUT_QUEUE_EMPTY), _fileSet(PortableStringHashFunc), _bitmap(NULL),
     _isConnected(false), _isConnecting(false), _wasConnected(false), _currentFileIndex(0),
     _localSessionID(localSessionID), _errorOccurred(false), _uploadSession(false),
     _origFileSetSize(0), _isAccepting(false), _isWaiting(false)
{
   // empty
}

ShareFileTransfer ::
~ShareFileTransfer()
{
   delete _mtt;
}

void 
ShareFileTransfer::
AddRequestedFileName(const char * fileName)
{
   _fileSet.Put(PortableString(fileName), true);
}

status_t 
ShareFileTransfer::
StartConnectSession(const char * hostName, int port, const char * remoteSessionID, bool wait)
{
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _origFileSetSize = _fileSet.Size();
   _uploadSession = (_origFileSetSize == 0);
   _isWaiting = wait;

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   status_t ret = _mtt->StartConnectThread(hostName, port, &_onConnect, &_onDisconnect);
   if (ret == B_NO_ERROR) _isConnecting = true;
   return ret;
}

status_t
ShareFileTransfer ::
StartAcceptSession(const char * remoteSessionID)
{
   _isAccepting = true;
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _origFileSetSize = _fileSet.Size();
   _uploadSession = (_origFileSetSize == 0);

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   status_t ret = _mtt->StartAcceptThread(0, false, &_onConnect, &_onDisconnect);
   if (ret == B_NO_ERROR) _isAccepting = true;
   return ret;
}


status_t 
ShareFileTransfer::
StartSocketUploadSession(int socket, bool wait)
{
   _isConnected = _uploadSession = true;
   _origFileSetSize = -1;  // we'll know this when we get the TRANSFER_COMMAND_FILE_LIST message
   _isWaiting = wait;

   BMessenger toMe(this);

   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   status_t ret = _mtt->StartThread(new PortableMessageIOGateway(new TCPSocketDataIO(socket, false)), socket, &_onDisconnect);
   if ((ret == B_NO_ERROR)&&(_isWaiting)) SendNotifyQueuedMessage();
   return ret;
}


void
ShareFileTransfer ::
SendNotifyQueuedMessage()
{
   PortableMessage * notifyMsg = _mtt->GetMessagePool()->GetObject();
   if (notifyMsg)
   {
      PortableMessageRef msgRef(notifyMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
      notifyMsg->what = TRANSFER_COMMAND_NOTIFY_QUEUED;
      _mtt->SendOutgoingMessage(msgRef);
   }
}



void 
ShareFileTransfer::
MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case TRANSFER_COMMAND_CONNECTED_TO_PEER:
         _isConnecting = false;
         _isConnected = _wasConnected = true;
         if (_mtt)  // paranoia
         {
            if (_uploadSession)
            {
               if (_isWaiting) SendNotifyQueuedMessage();
            }
            else
            {
               // Tell the remote BeShare connection what files we want from him
               PortableMessage * fileRequest = _mtt->GetMessagePool()->GetObject();
               if (fileRequest)
               {
                  PortableMessageRef fileRequestRef(fileRequest, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
                  fileRequest->what = TRANSFER_COMMAND_FILE_LIST;
                  fileRequest->AddString("beshare:FromSession", _localSessionID);
                  PortableHashtableIterator<PortableString, bool> iter = _fileSet.GetIterator();
                  PortableString * next;
                  while((next = iter.GetNextKey()) != NULL) fileRequest->AddString("files", next->Cstr());
                  _mtt->SendOutgoingMessage(fileRequestRef);
               }
               else AbortSession(true);
            }
         }
         ((ShareWindow*)Looper())->FileTransferConnected(this);
      break;

      case TRANSFER_COMMAND_DISCONNECTED_FROM_PEER:
         _isConnecting = _isConnected = _isAccepting = false;
         ((ShareWindow*)Looper())->FileTransferDisconnected(this);
         delete _mtt;  // might as well get rid of it, it's useless now
         _mtt = NULL;
      break;

      case TRANSFER_OUTPUT_QUEUE_EMPTY:
         if (_uploadSession) DoUpload();
      break;

      case PORTABLE_MESSAGES_RECEIVED:
      {
         // This BMessage comes from our MessageTransceiverThread, and
         // tells us that there may be some incoming PortableMessages to look at.
         if (_mtt)
         {
            PortableMessageRef next;
            while((_mtt)&&(_mtt->GetNextIncomingMessage(next) == B_NO_ERROR)) PortableMessageReceived(next.GetItemPointer());
         }
      }
      break;            
   }
}


void
ShareFileTransfer ::
DoUpload()
{
   if (_isWaiting) return;   // not yet!

   if (_mtt)
   {
      PortableMessage * msg = _mtt->GetMessagePool()->GetObject();
      if (msg)
      {
         PortableMessageRef msgRef(msg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
         if (_currentFileBytesDone >= 0)
         {
            ssize_t numBytes = _currentFile.Read(_scratchBuf, sizeof(_scratchBuf));
            if (numBytes > 0) 
            {
               msg->what = TRANSFER_COMMAND_FILE_DATA;
               msg->AddData("data", B_RAW_TYPE, _scratchBuf, numBytes);
               _mtt->SendOutgoingMessage(msgRef);
               _currentFileBytesDone += numBytes;
               ((ShareWindow *)Looper())->RefreshTransferItem(this);
               return;  
            }

            if (numBytes == 0)
            {
               // move to the next file
               _currentFileBytesDone = _currentFileSize = -1;
               DoUpload();
               return;
            }
         }
         else
         {
            // Try to get the next file from our send list
            if (_fileSetIter.GetNextKey(_currentFileName))
            {
               if ((_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_READ_ONLY) == B_NO_ERROR)&&
                   (_currentFile.GetSize(&_currentFileSize) == B_NO_ERROR))
               {
                  _currentFileIndex++;

                  if (((ShareWindow*)Looper())->LogUploads())
                  {
                     PortableString str("User ");
                     str += _remoteSessionID;
                     if (_remoteUserName[0])
                     {
                        str += " (a.k.a. ";
                        str += _remoteUserName;
                        str += ")";
                     }
                     str += " is downloading ";
                     str += _currentFileName;
                     ((ShareWindow*)Looper())->LogMessage(LOG_INFORMATION_MESSAGE, str.Cstr());
                  }

                  msg->AddString("beshare:File Name", _currentFileName.Cstr());
                  msg->AddInt64("beshare:File Size", _currentFileSize);
                  msg->AddString("beshare:FromSession", _localSessionID);

                  char attrName[B_ATTR_NAME_LENGTH];
                  BNodeInfo ni(&_currentFile);
                  if ((ni.InitCheck() == B_NO_ERROR)&&(ni.GetType(attrName) == B_NO_ERROR))
                  {
                     msg->AddString("beshare:Kind", attrName);
                     _bitmap = ((ShareWindow*)Looper())->GetBitmap(attrName);
                  }
                  else _bitmap = NULL;

                  while(_currentFile.GetNextAttrName(attrName) == B_NO_ERROR)
                  {
                     struct attr_info attrInfo;
                     if (_currentFile.GetAttrInfo(attrName, &attrInfo) == B_NO_ERROR)
                     {
                        char * attrData = new char[attrInfo.size];
                        ssize_t ret = _currentFile.ReadAttr(attrName, attrInfo.type, 0L, attrData, attrInfo.size);
                        if (ret == attrInfo.size) (void) msg->AddData(attrName, attrInfo.type, attrData, attrInfo.size);
                     }
                  }

                  ((ShareWindow*)Looper())->RefreshTransferItem(this);

                  msg->what = TRANSFER_COMMAND_FILE_HEADER;
                  _mtt->SendOutgoingMessage(msgRef);
                  _currentFileBytesDone = 0;               
                  DoUpload();
                  return;
               }
            }
            else
            {
               // All done!
               AbortSession(false);
               return;
            }
         }
      }
   }

   // If we got here, there was an error
   AbortSession(true);
}


void 
ShareFileTransfer ::
PortableMessageReceived(PortableMessage * msg)
{
   switch(msg->what)
   {
      case TRANSFER_COMMAND_NOTIFY_QUEUED:
         if (_uploadSession == false) 
         {
            _isWaiting = true;
            ((ShareWindow *)Looper())->RefreshTransferItem(this);
         }
      break;

         
      // Received when the remote peer wishes to tell us what files he wants to download from us
      case TRANSFER_COMMAND_FILE_LIST:
      {
         if (_uploadSession) 
         {
            if (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR) 
            {
               _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
            }
            else _remoteUserName = "Unknown";

            const char * nextFile;
            for (int i=0; (msg->FindString("files", i, &nextFile) == B_NO_ERROR); i++) AddRequestedFileName(nextFile);
            int numFiles = _fileSet.Size();
            if (numFiles > 0) 
            {
               _origFileSetSize = numFiles;
               _fileSetIter = _fileSet.GetIterator();
               DoUpload();
            }
            else AbortSession(true);   // he's gotta ask for something, otherwise why did he connect?
         }
         else AbortSession(true);  // you're going the wrong way buddy!
      }
      break;      

      // When downloading, this indicates the start of the next incoming file
      case TRANSFER_COMMAND_FILE_HEADER:
      {
         if ((_uploadSession == false)                                             &&
             (msg->FindString("beshare:File Name", _currentFileName) == B_NO_ERROR)&&
             (msg->FindInt64("beshare:File Size", &_currentFileSize) == B_NO_ERROR)&&
             (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR)&&
             (_fileSet.Remove(_currentFileName))                                   &&
             (GenerateNewFilename(_currentFileName) == B_NO_ERROR)                 &&
             (_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_WRITE_ONLY|B_CREATE_FILE) == B_NO_ERROR))
         {
            _isWaiting = false;  // obviously....
            _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
            _currentFileIndex++;
            _currentFileBytesDone = 0;
            PortableMessageFieldNameIterator iter = msg->GetFieldNameIterator();
            PortableString fieldName;
            while(iter.GetNextFieldName(fieldName))
            {
               if (strncmp(fieldName.Cstr(), "beshare:", 8))  // don't save our attributes, they aren't necessary
               {
                  const void * attrData;
                  size_t attrSize;
                  int32 c;
                  type_code type;
                  if ((msg->GetInfo(fieldName.Cstr(), &type, &c)                   == B_NO_ERROR)&&
                      (msg->FindData(fieldName.Cstr(), type, &attrData, &attrSize) == B_NO_ERROR))
                  {
                     (void)_currentFile.WriteAttr(fieldName.Cstr(), type, 0, attrData, attrSize);
                  }
               }
            }

            BNodeInfo ni(&_currentFile);
            const char * mimeString;
            if ((ni.InitCheck() == B_NO_ERROR)&&(msg->FindString("beshare:Kind", &mimeString) == B_NO_ERROR))
            {
               ni.SetType(mimeString);
               _bitmap = ((ShareWindow*)Looper())->GetBitmap(mimeString);
            }
            else _bitmap = NULL;

            ((ShareWindow*)Looper())->RefreshTransferItem(this);
         }
         else AbortSession(true);
      }
      break;

      case TRANSFER_COMMAND_FILE_DATA:
      {      
         const uint8 * data;
         size_t numBytes;
         if ((_uploadSession == false)&&(msg->FindData("data", B_RAW_TYPE, (const void **) &data, &numBytes) == B_NO_ERROR))
         {
            _isWaiting = false;  // obviously....
            if (_currentFile.Write(data, numBytes) == numBytes) 
            {
               _currentFileBytesDone += numBytes;
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
            }
            else 
            {
               PortableString errStr("Error writing to file ");
               errStr += _currentFileName;
               ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
               AbortSession(true);
            }
         }
         else AbortSession(true);
      }
      break;
   }
}

status_t
ShareFileTransfer ::
GenerateNewFilename(PortableString & str) const
{
   for (int i=0; i<50; i++)
   {
      BFile f(&_dir, str.Cstr(), B_READ_ONLY);
      if (f.InitCheck() != B_NO_ERROR) return B_NO_ERROR;

      int lastSpaceIdx = str.LastIndexOf(' ');
      if ((lastSpaceIdx >= 0)&&(lastSpaceIdx < str.Length()-1))
      {
         const char * cs = str.Cstr()+lastSpaceIdx+1;
         if ((cs[0] >= '0')&&(cs[0] <= '9'))
         {
            char buf[20];
            sprintf(buf, "%i", atoi(cs)+1);
            str = str.Substring(0, lastSpaceIdx+1) + buf;
            continue;
         }
      }
      str += " 2";  // default permutation
   }
   return B_ERROR;  // give up, something has gone horribly wrong
}


void
ShareFileTransfer::
AbortSession(bool error)
{
   if (_mtt)
   {
      ((ShareWindow*)Looper())->FileTransferDisconnected(this);
      if (error == false) _mtt->WaitForAllMessagesToBeSent();
      delete _mtt;
      _mtt = NULL;
      _isConnecting = _isAccepting = _isConnected = false;
   }
   _currentFileBytesDone = _currentFileSize = -1;
   _errorOccurred = error;
}


void 
ShareFileTransfer ::
Update(BView * lv, const BFont * font)
{
   BListItem::Update(lv, font);
   SetHeight(TRANSFER_ITEM_HEIGHT);
   font_height fontAttrs;
   font->GetHeight(&fontAttrs);
   _fontHeight = ceil(fontAttrs.ascent) + ceil(fontAttrs.descent);
}

void 
ShareFileTransfer ::
DrawItem(BView * lv, BRect itemRect, bool complete)
{
   rgb_color backgroundColor = ((_isConnected == false)&&(_isConnecting == false)) ? Grey : (_uploadSession ? PalePink : PaleBlue);
 
   lv->SetHighColor(backgroundColor);
   lv->FillRect(itemRect);

   BRegion Region;
   Region.Include(itemRect);
   lv->ConstrainClippingRegion(&Region);
   PortableString text;
   float percent = 0.0f;
   if (_isConnected)
   {
      if (_isWaiting) text = "Queued, please wait";
      else
      {
         text = _uploadSession ? "Sent: " : "Rcvd: ";

         if (_currentFileSize > 0)
         {
            char doneBuf[64], sizeBuf[64];
            GetByteSizeString(_currentFileBytesDone, doneBuf);
            GetByteSizeString(_currentFileSize, sizeBuf);
            char buf[100];
            sprintf(buf, "%s/%s", doneBuf, sizeBuf);
            text += buf;
            percent = ((float)_currentFileBytesDone)/_currentFileSize;
         }
      }
   }
   else
   {
      if (_errorOccurred) text = "An Error Occurred.";
      else
      {
         if (_isAccepting) text = "Awaiting callback...";
         else if (_isConnecting) text = "Connecting...";
         else
         {
            if (_uploadSession) text = "Preparing to Upload";
            else
            {
               if (_wasConnected) text = (_fileSet.Size() == 0) ? "Download complete" : "Download aborted.";
                             else text = "Couldn't connect";
            }
         }
      }
   }

   if (percent > 0.0f)
   {
      const rgb_color barColor = _uploadSession ? BrightPink : BrightBlue;
      BRect fillRect(itemRect);
      fillRect.right = (fillRect.left + fillRect.Width()*percent);
      lv->SetHighColor(barColor);
      lv->FillRect(fillRect);
      if (percent > 0.5f) backgroundColor = barColor;  // do text background from whoever has majority
   }

   lv->SetLowColor(backgroundColor);
   lv->SetHighColor(Black);

   char line2[20];
   sprintf(line2, "(%i/%i) %s %s", _currentFileIndex, _origFileSetSize, _uploadSession?"to":"from", _remoteUserName.Cstr());
   lv->DrawString(line2, BPoint(2.0, itemRect.bottom-(_fontHeight+5.0f)));

   lv->DrawString(_currentFileName.Cstr(), BPoint((_bitmap?_bitmap->Bounds().Width()+6.0f:2.0f), itemRect.top+2.0f+_fontHeight));
   lv->DrawString(text.Cstr(), BPoint(itemRect.left+2.0f,itemRect.bottom-4.0f));
   lv->ConstrainClippingRegion(NULL);
   if (_bitmap)
   {
      lv->SetDrawingMode(B_OP_OVER);
      lv->DrawBitmap(_bitmap, BPoint(itemRect.left + 2.0f, itemRect.top+2.0f));
   }

   // Do border highlighting
   bool selected = IsSelected();
   lv->SetHighColor(selected ? Black : White);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.right, itemRect.top));
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.left, itemRect.bottom));
   lv->SetHighColor(selected ? White : Black);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.bottom), BPoint(itemRect.right, itemRect.bottom));
   lv->StrokeLine(BPoint(itemRect.right, itemRect.top), BPoint(itemRect.right, itemRect.bottom));
}



void
ShareFileTransfer ::
BeginTransfer()
{
   if (_isWaiting)
   {
      _isWaiting = false;
      if ((_uploadSession)&&(_isConnected)&&(_fileSet.Size() > 0)) DoUpload();
   }
}
