#ifndef SHARE_FILE_TRANSFER_H
#define SHARE_FILE_TRANSFER_H

#include <app/Handler.h>
#include <storage/File.h>
#include <storage/Directory.h>
#include <interface/Bitmap.h>
#include <interface/Font.h>
#include <interface/ListItem.h>

#include "hashtable/PortableHashtable.h"
#include "string/PortableString.h"
#include "besupport/MessageTransceiverThread.h"

#define TRANSFER_ITEM_HEIGHT 47.0f 

/*
 * This class handles the sending and receiving of one or more files 
 * to/from another BeShare client.  It also serves as a BListItem 
 * for the BeShare GUI, so that the user can track the download's
 * progress.
 */
class ShareFileTransfer : public BHandler, public BListItem 
{
public:
   /* 
    * Creates a new session.  Call AddFile() on it with whatever files you 
    * want it to access for upload/download, then call Start*Session().
    */
   ShareFileTransfer(const BDirectory & fileDir, const char * localSessionID);

   /*
    * Destroys the ShareFileTransfer object, cancelling any transfer currently in progress. 
    */
   ~ShareFileTransfer();

   /* 
    * Adds the given file to our list of files to request.
    */
   void AddRequestedFileName(const char * fileName);

   /*
    * Initiates an outgoing connection to (hostName, port).  Once connected, behaviour is
    * determined by our file set:  If we have no files to request, we will allow the remote
    * user to request and download files from us, otherwise we will request and download files
    * from him.
    * If (wait) is true, file uploading won't start until our BeginTransfer() method is called.
    */
   status_t StartConnectSession(const char * hostName, int port, const char * remoteSessionID, bool wait);

   /*
    * Begins listening on a chosen port for incoming connections.  When a connection is received,
    * handle it as described above (in StartConnectSession())
    */
   status_t StartAcceptSession(const char * remoteSessionID);

   /*
    * Returns the port we are listening on for an incoming connection, or -1 if we aren't listening.
    */
   int GetPort() const {return _mtt ? _mtt->GetPort() : -1;}

   /*
    * Begins a session that will allow the remote user to request files to be sent to him.
    * (beginImmediately) specifies whether to start sending data ASAP, or to wait until 
    * the ShareWindow tells us to.
    */
   status_t StartSocketUploadSession(int socket, bool beginImmediately);

   /*
    * Handles BMessages from the MessageTransceiverThread 
    */
   virtual void MessageReceived(BMessage * msg);

   /* 
    * Handles PortableMessages from the MessageTransceiverThread
    */
   void PortableMessageReceived(PortableMessage * msg);

   /* BListItem interface */
   virtual void Update(BView * owner, const BFont * font);
   virtual void DrawItem(BView * owner, BRect itemRect, bool complete = false);

   /* Called if we want the list item to show a MIME file icon.
    * Note that this object does not assume ownership of (bmap).
    */
   void SetBitmap(const BBitmap * bmap);

   /* Returns our current set of files to be uploaded or downloaded */
   PortableHashtable<PortableString, bool> & GetFileSet() {return _fileSet;}

   /* Accessors for some miscellaneous state variables */
   bool IsUploadSession() const {return _uploadSession;}
   bool IsConnected() const {return _isConnected;}
   bool IsConnecting() const {return _isConnecting;}
   bool WasConnecting() const {return _wasConnected;}
   bool IsWaiting() const {return _isWaiting;}

   /* For upload sessions that were previously asked to wait, this will clear the
    * wait flag and begin the upload.
    */
   void BeginTransfer();

private:
   status_t GenerateNewFilename(PortableString & getSetFilename) const;
   void DoUpload();
   void AbortSession(bool errorOccurred);
   void SendNotifyQueuedMessage();

   enum 
   {
      TRANSFER_COMMAND_CONNECTED_TO_PEER='tshr', 
      TRANSFER_COMMAND_DISCONNECTED_FROM_PEER,
      TRANSFER_COMMAND_FILE_LIST,     // a list of files the remote user would like to download
      TRANSFER_COMMAND_FILE_HEADER,   // contains filename, attributes, etc.
      TRANSFER_COMMAND_FILE_DATA,     // a chunk of the file's contents
      TRANSFER_OUTPUT_QUEUE_EMPTY,    // sent by MessageTransceiverThread when it needs more data
      TRANSFER_COMMAND_NOTIFY_QUEUED  // tells the receiving session he's being put on a wait list to download
   };

   const BDirectory & _dir;

   PortableHashtable<PortableString, bool> _fileSet;  // set of files to request from remote peer
   PortableHashtableIterator<PortableString, bool> _fileSetIter;            // keeps track of where we are in our fileset for uploading

   BFile _currentFile;
   PortableString _currentFileName;
   off_t _currentFileSize;
   off_t _currentFileBytesDone;

   MessageTransceiverThread * _mtt;

   bool _uploadSession;
 
   BMessage _onConnect;
   BMessage _onDisconnect;
   BMessage _onEmpty;

   const BBitmap * _bitmap;
   float _fontHeight;

   // Some state vars to govern our GUI display
   PortableString _localSessionID;
   PortableString _remoteSessionID;
   PortableString _remoteUserName;
   bool _isAccepting;
   bool _isConnecting;
   bool _isConnected;
   bool _isWaiting;   // true iff we've been queued
   bool _wasConnected;
   bool _errorOccurred;
   int _currentFileIndex;
   int _origFileSetSize;

   uint8 _scratchBuf[8192];  // temp area
};

#endif
