#include <unistd.h>
#include <net/netdb.h>
#include <kernel/fs_attr.h>
#include <storage/Mime.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/NodeInfo.h>
#include <storage/Path.h>

#include "besupport/ConvertMessages.h"
#include "reflector/StorageReflectConstants.h"
#include "reflector/StorageReflectUtils.h"
#include "netutil/NetworkUtilityFunctions.h"

#include "ShareConstants.h"
#include "ShareNetClient.h"
#include "ShareWindow.h"


static uint32 InoTHashFunc(const ino_t & key);
static uint32 InoTHashFunc(const ino_t & key)
{
   return (uint32) (key % ((uint32)-1));  // hash 64-bit int to a 32-bit int
}

ShareNetClient ::
ShareNetClient(const BDirectory & shareDir, int32 localPort) :
   _localUserName("Anonymous"), _shareDir(shareDir), _mtt(NULL), _nodeToEntry(InoTHashFunc), _localSharePort(localPort), _firewalled(false), _bandwidth(0), _bandwidthLabel("Unknown")
{
   // empty
}


ShareNetClient ::
~ShareNetClient()
{
   DisconnectFromServer();
}


void
ShareNetClient ::
ConnectToServer(const char * serverName, int port)
{
   DisconnectFromServer();  // clean out any old connection, as we don't want two at once.

   BMessage sendOnConnect(NET_CLIENT_CONNECTED_TO_SERVER);
   BMessage sendOnDisconnect(NET_CLIENT_DISCONNECTED_FROM_SERVER);

   BMessenger toMe(this);
   _mtt = new MessageTransceiverThread(toMe);
   if (_mtt->StartConnectThread(serverName, port, &sendOnConnect, &sendOnDisconnect) == B_NO_ERROR)
   {
      ((ShareWindow*)Looper())->SetConnectStatus(true, false);
   }
   else
   {
      ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, "Error, couldn't create connect thread!");
      delete _mtt;
      _mtt = NULL;
   }
}


void
ShareNetClient ::
DisconnectFromServer()
{
   if (_mtt)
   {
      delete _mtt;
      _mtt = NULL;

      // since we're no longer connected, there's no point in watching the shared dir anymore
      node_ref shareNodeRef;
      (void)_shareDir.GetNodeRef(&shareNodeRef);
      (void)watch_node(&shareNodeRef, B_STOP_WATCHING, this);
      _nodeToEntry.Clear();

      ((ShareWindow *)Looper())->SetConnectStatus(false, false);
   }
}


void
ShareNetClient ::
StartQuery(const char * sessionIDRegExp, const char * fileNameRegExp)
{
   if (_mtt)
   {
      StopQuery();          // get rid of the currently running query, if any

      // This path string tells muscled which files it should inform us about.
      // Since we may be basing part of our query on the session ID, we use
      // the full path string for this query.
      PortableString temp = "SUBSCRIBE:/*/";
      temp += sessionIDRegExp;
      temp += "/beshare/";
      temp += _firewalled ? "files/" : "fi*/";  // If we're firewalled, we can only get non-firewalled files; else both types
      temp += fileNameRegExp;
      AddServerSubscription(temp.Cstr());
   }
}


void
ShareNetClient ::
StopQuery()
{
   // Instead of having to remember what the exact subscription string
   // was, it's easier to just tell the muscle server to cancel all
   // subscriptions that have the substring 'beshare/fi' in them.
   if (_mtt) RemoveServerSubscription("SUBSCRIBE:*beshare/fi*");
}


void
ShareNetClient ::
MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case NET_CLIENT_CONNECTED_TO_SERVER:
      {
         // Ask the server to send us our parameters.  The parameters
         // message it will send contains our machine's hostname and session
         // ID, amongst other things (we'll need to know those later)
         PortableMessage * pmsg = _mtt->GetMessagePool()->GetObject();
         if (pmsg)
         {
            PortableMessageRef pmsgRef(pmsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
            pmsg->what = PR_COMMAND_GETPARAMETERS;
            _mtt->SendOutgoingMessage(pmsgRef);
         }

         ((ShareWindow *)Looper())->SetConnectStatus(true, true);
         BeginSharingFiles();
      }
      break;

      case NET_CLIENT_DISCONNECTED_FROM_SERVER:
         DisconnectFromServer();  // clean up the dead MessageTransceiverThread object
      break;

      case PORTABLE_MESSAGES_RECEIVED:
      {
         // This BMessage comes from our MessageTransceiverThread, and 
         // tells us that there may be some incoming PortableMessages to look at.  
         PortableMessageRef next;
         while(_mtt->GetNextIncomingMessage(next) == B_NO_ERROR) PortableMessageReceived(next.GetItemPointer());
      }
      break;

      case B_NODE_MONITOR:   // something in our share directory has changed!
      {
         int32 opCode;
         node_ref fileNodeRef;
         if ((msg->FindInt32("opcode", &opCode)             == B_NO_ERROR)&&
             (msg->FindInt64("node",   &fileNodeRef.node)   == B_NO_ERROR)&&
             (msg->FindInt32("device", &fileNodeRef.device) == B_NO_ERROR))
         {
            switch(opCode)
            {
               case B_ENTRY_CREATED:
               {
                  const char * name;
                  ino_t directory;
                  if ((msg->FindInt64("directory", &directory) == B_NO_ERROR) &&
                      (msg->FindString("name", &name)          == B_NO_ERROR))
                  {
                     NodeCreated(fileNodeRef, entry_ref(fileNodeRef.device, directory, name));
                  }
               }
               break;

               case B_ENTRY_MOVED:
               {
                  NodeRemoved(fileNodeRef.node);  // Out with the old (if any)...
                    
                  // And if the new file is in our dir, (re)add it.
                  node_ref shareDirNode;
                  ino_t newDirNode;
                  const char * name;
                  if ((msg->FindInt64("to directory", &newDirNode) == B_NO_ERROR)&&
                      (_shareDir.GetNodeRef(&shareDirNode)         == B_NO_ERROR)&&
                      (shareDirNode.node == newDirNode)                          &&
                      (msg->FindString("name", &name)              == B_NO_ERROR))
                  {
                     NodeCreated(fileNodeRef, entry_ref(fileNodeRef.device, newDirNode, name));
                  }
               }
               break;

               case B_ENTRY_REMOVED:
                  NodeRemoved(fileNodeRef.node);
               break;
            }
         }
      }
      break;

      default:
         BHandler::MessageReceived(msg);
      break;
   }
}

void
ShareNetClient ::
NodeRemoved(ino_t node)
{
   entry_ref ref;
   if ((_nodeToEntry.Remove(node, ref))&&(_mtt))
   {
      PortableMessage * uploadMsg = _mtt->GetMessagePool()->GetObject();
      if (uploadMsg)
      {
         uploadMsg->what = PR_COMMAND_REMOVEDATA;
         PortableMessageRef uploadMsgRef(uploadMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());

         PortableString path(GetFilesNodePath());
         path += ref.name;
         EscapeRegexTokens(path);
         uploadMsg->AddString(PR_NAME_KEYS, path.Cstr());
         _mtt->SendOutgoingMessage(uploadMsgRef);
      }
   }
}

void 
ShareNetClient ::
NodeCreated(const node_ref & noderef, const entry_ref & entryref)
{
   PortableMessageRef value;
   if (GetFileInfo(value, entryref, noderef) == B_NO_ERROR)
   {
      _nodeToEntry.Put(noderef.node, entryref);
      PortableString nodepath(GetFilesNodePath());
      nodepath += entryref.name;
      SetDataNodeValue(nodepath.Cstr(), value);
   }
}

void
ShareNetClient ::
PortableMessageReceived(PortableMessage * msg)
{
   switch(msg->what)
   {
      case NET_CLIENT_CONNECT_BACK_REQUEST:
      {
         const char * session;
         int32 port;
         if ((_firewalled) &&  // not firewalled?  Then they should just connect to us directly
             (msg->FindString("session", &session) == B_NO_ERROR) &&
             (msg->FindInt32("port", &port)        == B_NO_ERROR)) ((ShareWindow*)Looper())->ConnectBackRequestReceived(session, port);
      }
      break;
 
      case NET_CLIENT_NEW_CHAT_TEXT:
      {
         // Someone has sent a line of chat text to display
         const char * text;
         const char * session;
         const rgb_color privColor = {0, 150, 150, 255};
         if ((msg->FindString("text", &text) == B_NO_ERROR)&&(msg->FindString("session", &session) == B_NO_ERROR)) ((ShareWindow*)Looper())->LogMessage(LOG_REMOTE_USER_CHAT_MESSAGE, text, session, msg->HasName("private") ? &privColor : NULL);
      }
      break;

      case PR_RESULT_DATAITEMS:
      {
         // Part of the server-side database that we subscribed to has changed
         ((ShareWindow*)Looper())->BeginBatchFileResultUpdate();

         // Look for sub-messages that indicate that nodes were removed from the tree
         PortableString nodepath;
         for (int i=0; (msg->FindString(PR_NAME_REMOVED_DATAITEMS, i, nodepath) == B_NO_ERROR); i++) 
         {
            int pathDepth = GetPathDepth(nodepath.Cstr());
            if (pathDepth >= USER_NAME_DEPTH)
            {
               PortableString sessionID = GetPathClause(SESSION_ID_DEPTH, nodepath.Cstr());
               sessionID = sessionID.Substring(0, sessionID.IndexOf('/'));

               switch(GetPathDepth(nodepath.Cstr()))
               {
                  case USER_NAME_DEPTH: 
                     if (strncmp(GetPathClause(USER_NAME_DEPTH, nodepath.Cstr()), "name", 4) == 0) ((ShareWindow*)Looper())->RemoveUser(sessionID.Cstr());
                  break;

                  case FILE_INFO_DEPTH: 
                     ((ShareWindow*)Looper())->RemoveResult(sessionID.Cstr(), GetPathClause(FILE_INFO_DEPTH, nodepath.Cstr())); 
                  break;
               }
            }
         }

         // Look for sub-messages that indicate that nodes were added to the tree
         PortableMessageFieldNameIterator iter = msg->GetFieldNameIterator(B_MESSAGE_TYPE);
         while(iter.GetNextFieldName(nodepath))
         {
            int pathDepth = GetPathDepth(nodepath.Cstr());
            if (pathDepth >= USER_NAME_DEPTH)
            {
               PortableMessageRef tempRef;
               if (msg->FindMessage(nodepath.Cstr(), tempRef) == B_NO_ERROR)
               {
                  PortableString sessionID = GetPathClause(SESSION_ID_DEPTH, nodepath.Cstr());
                  sessionID = sessionID.Substring(0, sessionID.IndexOf('/'));
                  switch(pathDepth)
                  {
                     case USER_NAME_DEPTH: 
                     {
                        PortableString hostName = GetPathClause(HOST_NAME_DEPTH, nodepath.Cstr());
                        hostName = hostName.Substring(0, hostName.IndexOf('/'));

                        const char * nodeName = GetPathClause(USER_NAME_DEPTH, nodepath.Cstr());
                        if (strncmp(nodeName, "name", 4) == 0)
                        {
                           int32 port;
                           const char * name;
                           if ((tempRef.GetItemPointer()->FindString("name", &name) == B_NO_ERROR)&& 
                               (tempRef.GetItemPointer()->FindInt32("port", &port) == B_NO_ERROR))
                              ((ShareWindow *)Looper())->PutUser(sessionID.Cstr(), name, hostName.Cstr(), port);
                        }
                        else if (strncmp(nodeName, "bandwidth", 9) == 0)
                        {
                           uint32 bps;
                           const char * label;
                           if ((tempRef.GetItemPointer()->FindString("label", &label) == B_NO_ERROR)&& 
                               (tempRef.GetItemPointer()->FindInt32("bps", (int32*)&bps) == B_NO_ERROR))
                              ((ShareWindow *)Looper())->SetUserBandwidth(sessionID.Cstr(), label, bps);
                        }
                     }
                     break;

                     case FILE_INFO_DEPTH: 
                     {
                        ((ShareWindow *)Looper())->PutResult(sessionID.Cstr(), 
GetPathClause(FILE_INFO_DEPTH, nodepath.Cstr()), (GetPathClause(USER_NAME_DEPTH, nodepath.Cstr())[2] == 'r'), tempRef);
                     }
                     break;
                  }
               }
            }
         }

         ((ShareWindow*)Looper())->EndBatchFileResultUpdate();
      }
      break;

      case PR_RESULT_PARAMETERS:
      {
         const char * sessionRoot;
         if (msg->FindString(PR_NAME_SESSION_ROOT, &sessionRoot) == B_NO_ERROR)
         {
            // session root is of form "/hostname/sessionID"; parse these out
            const char * lastSlash = strrchr(sessionRoot, '/');
            if (lastSlash)
            {
               _localSessionID = lastSlash+1;
               _localHostName = PortableString(sessionRoot).Substring(1, (lastSlash-sessionRoot));
            }
         }
      }
      break;
   }
}




const char * 
ShareNetClient :: GetFilesNodePath() const
{
   return _firewalled ? "beshare/fires/" : "beshare/files/";
}


void
ShareNetClient ::
BeginSharingFiles()
{
   if (_mtt)
   {
      // upload our user name...
      SetLocalUserName(_localUserName.Cstr());

      if (_localSharePort > 0)
      {
         // Here we upload the file info node values for *all* our shared
         // files in just one PR_COMMAND_SETDATA message.
         node_ref shareNodeRef;
         if ((_shareDir.GetNodeRef(&shareNodeRef) == B_NO_ERROR)&& 
             (watch_node(&shareNodeRef, B_WATCH_DIRECTORY, this) == B_NO_ERROR))
         {
            if (_shareDir.CountEntries() > 0)
            {
               char buf[100];
               sprintf(buf, "(Sharing %li local files)", _shareDir.CountEntries());
               ((ShareWindow*)Looper())->LogMessage(LOG_INFORMATION_MESSAGE, buf);

               PortableMessage * uploadMsg = _mtt->GetMessagePool()->GetObject();
               if (uploadMsg)
               {
                  uploadMsg->what = PR_COMMAND_SETDATA;
                  PortableMessageRef uploadMsgRef(uploadMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());

                  // scan directory, and for each file add a sub-message to our setdata message
                  entry_ref entryRef;
                  while(_shareDir.GetNextRef(&entryRef) == B_NO_ERROR) 
                  {
                     BNode node(&entryRef);
                     if (node.InitCheck() == B_NO_ERROR)
                     {
                        node_ref nodeRef;
                        PortableMessageRef nextFileRef;
                        if ((node.GetNodeRef(&nodeRef)                   == B_NO_ERROR)&&
                            (GetFileInfo(nextFileRef, entryRef, nodeRef) == B_NO_ERROR))
                        {
                           _nodeToEntry.Put(nodeRef.node, entryRef);
                           PortableString nextpath(GetFilesNodePath());

                           nextpath += entryRef.name;
                           uploadMsg->AddMessage(nextpath.Cstr(), nextFileRef);
                        }
                     }
                  }
                  _shareDir.Rewind();
                  _mtt->SendOutgoingMessage(uploadMsgRef);
               }
            }
         }
         else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, "Error, couldn't watch the share directory!");
      }

      // This subscription will let us know what user-names are present.
      AddServerSubscription("SUBSCRIBE:beshare/name");

      // And we'll want to know the guy's connection bandwidth too
      AddServerSubscription("SUBSCRIBE:beshare/bandwidth");
   }
}


status_t
ShareNetClient ::
GetFileInfo(PortableMessageRef & returnValue, const entry_ref & entryRef, const node_ref & nodeRef)
{
   BEntry entry(&entryRef);
   if (entry.InitCheck() == B_NO_ERROR)
   {
      off_t mainSize;
      BFile file(&entryRef, B_READ_ONLY);
      if ((file.InitCheck() == B_NO_ERROR)&&(file.GetSize(&mainSize) == B_NO_ERROR)) 
      {
         PortableMessage * infoMessage = _mtt->GetMessagePool()->GetObject();
         if (infoMessage)
         {
            returnValue = PortableMessageRef(infoMessage, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
            infoMessage->AddInt64("beshare:File Size", mainSize);
   
            // Now load the attributes for our file's MIME type and add them to our message
            BNodeInfo ni(&file);
            char mimeType[B_ATTR_NAME_LENGTH];
            if ((ni.InitCheck() == B_NO_ERROR)&&(ni.GetType(mimeType) == B_NO_ERROR))
            {
               infoMessage->AddString("beshare:Kind", mimeType);
               BMimeType mt(mimeType);
               if (mt.InitCheck() == B_NO_ERROR)
               {
                  BMessage attrInfo;
                  if (mt.GetAttrInfo(&attrInfo) == B_NO_ERROR)
                  {
                     const char * attrName;
                     for (int i=0; attrInfo.FindString("attr:name", i, &attrName) == B_NO_ERROR; i++)
                     {
                        type_code tc;
                        if (attrInfo.FindInt32("attr:type", (int32 *) &tc) == B_NO_ERROR)
                        {
                           uint8 attrData[256]; 
                           ssize_t ret = file.ReadAttr(attrName, tc, 0L, attrData, sizeof(attrData));
                           if (ret > 0) (void) infoMessage->AddData(attrName, tc, attrData, ret);
                        }
                     }
                  }
               }
            }
            return B_NO_ERROR;
         }
      }
   }
   return B_ERROR;
}


void 
ShareNetClient ::
AddServerSubscription(const char * subscriptionString, bool quietly)
{
   PortableMessage * queryMsg = _mtt->GetMessagePool()->GetObject();
   if (queryMsg)
   {
      queryMsg->what = PR_COMMAND_SETPARAMETERS;
      queryMsg->AddBool(subscriptionString, true);  // the true doesn't signify anything
     
      if (quietly) queryMsg->AddBool(PR_NAME_SUBSCRIBE_QUIETLY, true);  // suppress initial-state response

      // This PortableMessageRef object ensures that (queryMsg) gets recycled after it's sent.
      PortableMessageRef queryMsgRef(queryMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
      _mtt->SendOutgoingMessage(queryMsgRef);
   }
   else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, "AddServerSubscription:  Error, couldn't obtain message; out of memory?");
}


void
ShareNetClient ::
RemoveServerSubscription(const char * pattern)
{
   // Tell the server we're no longer interested in file info.
   PortableMessage * removeMsg = _mtt->GetMessagePool()->GetObject();
   if (removeMsg)
   {
      removeMsg->what = PR_COMMAND_REMOVEPARAMETERS;
      removeMsg->AddString(PR_NAME_KEYS, pattern);

      // This PortableMessageRef object ensures that (removeMsg) gets recycled after it's sent.
      PortableMessageRef removeMsgRef(removeMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
      _mtt->SendOutgoingMessage(removeMsgRef);
   }
   else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, "Error, couldn't obtain message; out of memory?");
}




void
ShareNetClient ::
SetLocalUserName(const char * name)
{
   _localUserName = name;
   if (_mtt)
   {
      PortableMessage * nameUploadMessage = _mtt->GetMessagePool()->GetObject();
      if (nameUploadMessage)
      {
         PortableMessage * nameMessage = _mtt->GetMessagePool()->GetObject();
         if (nameMessage)
         {
            nameMessage->AddString("name", _localUserName.Cstr());
            nameMessage->AddInt32("port", _localSharePort);
            SetDataNodeValue("beshare/name", PortableMessageRef(nameMessage, _mtt->GetMessagePool(), _mtt->GetMessageRefPool()));

            // Make sure server knows our bandwidth info
            SetUploadBandwidth(_bandwidthLabel.Cstr(), _bandwidth);
         }
      } 
   }
}


void
ShareNetClient ::
SendChatMessage(const char * targetSessionID, const char * messageText)
{
   if (_mtt)
   {
      PortableMessage * chatMessage = _mtt->GetMessagePool()->GetObject();
      if (chatMessage)
      {
         chatMessage->what = NET_CLIENT_NEW_CHAT_TEXT;

         PortableString toString("/*/");  // send message to all hosts...
         toString += targetSessionID;     // who have the given sessionID (or "*" == all session IDs)
         toString += "/beshare";          // and are beshare sessions.
         chatMessage->AddString(PR_NAME_KEYS, toString.Cstr());

         chatMessage->AddString("session", _localSessionID.Cstr());
         chatMessage->AddString("text", messageText);
         if (strcmp(targetSessionID, "*")) chatMessage->AddBool("private", true);
         _mtt->SendOutgoingMessage(PortableMessageRef(chatMessage, _mtt->GetMessagePool(), _mtt->GetMessageRefPool()));
      }
   }
}

void
ShareNetClient ::
SetDataNodeValue(const char * nodePath, PortableMessageRef nodeValue)
{
   if (_mtt)
   {
      PortableMessage * uploadMsg = _mtt->GetMessagePool()->GetObject();
      if (uploadMsg)
      {
         PortableMessageRef uploadMsgRef(uploadMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
         uploadMsg->what = PR_COMMAND_SETDATA;
         uploadMsg->AddMessage(nodePath, nodeValue);
         _mtt->SendOutgoingMessage(uploadMsgRef);
      }
   }
}

void
ShareNetClient ::
SendConnectBackRequestMessage(const char * targetSessionID, int port)
{
   if (_mtt)
   {
      PortableMessage * reqMsg = _mtt->GetMessagePool()->GetObject();
      if (reqMsg)
      {
         PortableMessageRef reqMsgRef(reqMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
         reqMsg->what = NET_CLIENT_CONNECT_BACK_REQUEST;

         PortableString toString("/*/");  // send message to all hosts...
         toString += targetSessionID;     // who have the given sessionID (or "*" == all session IDs)
         toString += "/beshare";          // and are beshare sessions.
         reqMsg->AddString(PR_NAME_KEYS, toString);
         reqMsg->AddString("session", _localSessionID);
         reqMsg->AddInt32("port", port);

         _mtt->SendOutgoingMessage(reqMsgRef);
      }
   }
}


void
ShareNetClient ::
SetUploadBandwidth(const char * l, uint32 bps)
{
   _bandwidth = bps;
   _bandwidthLabel = l;
   if (_mtt)
   {
      PortableMessage * msg = _mtt->GetMessagePool()->GetObject();
      if (msg)
      {
         PortableMessageRef value(msg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
         msg->AddString("label", l);
         msg->AddInt32("bps", bps);
         SetDataNodeValue("beshare/bandwidth", value);
      }
   }
}
