#ifndef SHARE_NET_CLIENT_H
#define SHARE_NET_CLIENT_H

#include <app/Handler.h>
#include <storage/NodeMonitor.h>
#include <storage/Directory.h>

#include "besupport/MessageTransceiverThread.h"
#include "hashtable/PortableHashtable.h"

/*
 * This class handles the TCP connection to the MUSCLE server, and
 * also maintains files in the shares (uploads) directory as well
 * as the query-results directory.
 */
class ShareNetClient : public BHandler 
{
public:
   /*
    * (shareDir) is the directory whose files we will make available to other users.
    * (localSharePort) is the port on the local machine that other BeShare clients
    * can connect to in order to download files from us.  It can be -1 if we
    * aren't connect-to-able.
    */
   ShareNetClient(const BDirectory & shareDir, int32 localSharePort);
   ~ShareNetClient();

   /* Closes any current connection to the MUSCLE server and starts a new one. */
   void ConnectToServer(const char * serverName, int port);

   /* Closes any current ocnnection to the MUSCLE server. */
   void DisconnectFromServer();

   /* Removes any current query, clears the current result set, and starts a new query. */
   void StartQuery(const char * sessionExp, const char * fileExp);

   /* Stops the current query, but doesn't clear the query directory. */
   void StopQuery();

   /* Changes the user's handle */
   void SetLocalUserName(const char * newName);

   /* Changes the local user's advertised upload bandwidth */
   void SetUploadBandwidth(const char * label, uint32 bps);

   /* Accessors */
   const char * GetLocalUserName() {return _localUserName.Cstr();}
   const char * GetLocalHostName() {return _localHostName.Cstr();}
   const char * GetLocalSessionID() {return _localSessionID.Cstr();}

   /* Uploads a chat message to server, to the given sessionID 
    * (a sessionID of '*' will broadcast to all beshare sessions)
    */
   void SendChatMessage(const char * targetSessionID, const char * text);

   /* Sends a message to the specified session asking him to connect back to
    * us at the specified port so we can download some files from him.  
    */
   void SendConnectBackRequestMessage(const char * targetSessionID, int port);

   /* Handles BMessages sent to us by the MessageTransceiverThread object. */
   virtual void MessageReceived(BMessage *message); 

   /* Get/set for whether or not we are behind a firewall.  SetFirewall()
    * should not be called when we are connected to the server */
   void SetFirewalled(bool f) {_firewalled = f;}
   bool GetFirewalled() const {return _firewalled;}

private:
   // Our server-side-directory node paths have one of the following forms:
   //   /<hostname>/<sessionid>/beshare/files/<filename>     (shared file info)
   //   /<hostname>/<sessionid>/beshare/name                 (user handle info)
   // The following enumeration gives meaningful names to the levels of the tree
   enum 
   {
      ROOT_DEPTH = 0,         // root node
      HOST_NAME_DEPTH,         
      SESSION_ID_DEPTH,
      BESHARE_HOME_DEPTH,     // used to separate our stuff from other (non-BeShare) data on the same server
      USER_NAME_DEPTH,        // user's handle node would be found here
      FILE_INFO_DEPTH         // user's shared file list is here
   };

   // 'what' codes that we use internally
   enum 
   {
      NET_CLIENT_CONNECTED_TO_SERVER = 0,
      NET_CLIENT_DISCONNECTED_FROM_SERVER,
      NET_CLIENT_NEW_CHAT_TEXT,
      NET_CLIENT_CONNECT_BACK_REQUEST
   }; 

   /* Called by MessageReceived, once per incoming PortableMessage */
   void PortableMessageReceived(PortableMessage * pmsg);

   /* Sends an error message to the ShareWindow to have it print a message into the chat window */
   void LogError(const char * msg);

   /* Uploads info on the contents of our shareDir to the server, and starts node-monitoring
    * the share dir for changes.
    */
   void BeginSharingFiles();

   /* Returns a PortableMessageRef that contains the shared data info for the given file. */
   status_t GetFileInfo(PortableMessageRef & returnRef, const entry_ref & entryRef, const node_ref & nodeRef);

   /* Convenience method; adds a data-node subscription to the server */
   void AddServerSubscription(const char * subString, bool quietly = false);

   /* Convenience method; removes one or more data-node subscriptions from the server. */
   void RemoveServerSubscription(const char * subStringRegExp);

   /* Convenience method: uploads the given value for the given node on the server */
   void SetDataNodeValue(const char * nodepath, PortableMessageRef nodeValue);

   /* Called when a new entry is discovered in our shared dir */
   void NodeCreated(const node_ref & nodeRef, const entry_ref & entryRef);

   /* Called when an entry is removed from our shared dir */
   void NodeRemoved(ino_t nodeNumber);

   /* Returns "/beshare/files" or "/beshare/fires" depending on whether we're firewalled. */
   const char * GetFilesNodePath() const;
   PortableString _localUserName;

   MessageTransceiverThread * _mtt;

   PortableHashtable<ino_t, entry_ref> _nodeToEntry;

   BDirectory _shareDir;

   // These are set when we get a SETPARAMS reply message
   PortableString _localHostName;
   PortableString _localSessionID;

   int32 _localSharePort;

   bool _firewalled;
  
   PortableString _bandwidthLabel;
   uint32 _bandwidth;
};

#endif
